/* SPDX-License-Identifier: GPL-2.0+
 *
 * linux/fs/erofs/pagevec.h
 *
 * Copyright (c) 2018 HUAWEI, Inc.
 *             http://www.huawei.com/
 * Created by Gao Xiang <gaoxiang25@huawei.com>
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file COPYING in the main directory of the Linux
 * distribution for more details.
 */
#ifndef __EROFS_PAGEVEC_H
#define __EROFS_PAGEVEC_H

#include "internal.h"

/*
 * page_start
 * |  /
 * | /
 * |/
 * +------+--------------------------------+-------+
 * | prev | (*1) |       ...               | *next |
 * +------+--------------------------------+-------+
 * NOTE: (*1) could be the pagevec page itself.
 */

/* used for traversaling zipped page in reserve order */
struct z_erofs_zipped_pagevec {
	struct page *page;
	struct page **pages;
};

static inline void z_erofs_zipped_pagevec_push(
	struct z_erofs_zipped_pagevec *vec,
	struct page *page)
{
	if (unlikely(vec->page == NULL)) {
		vec->page = page;
		vec->pages = kmap(page);
		*vec->pages = NULL;
	} else if (!offset_in_page(vec->pages + 1)) {
		struct page *const prev = vec->page;

		vec->page = *vec->pages;
		kunmap(prev);
		vec->pages = kmap(vec->page);
		*vec->pages = prev;
	}
	*++vec->pages = page;
}

#define z_erofs_zipped_pagevec_empty(v)	((v)->page == NULL)

static inline struct page *z_erofs_zipped_pagevec_pop(
	struct z_erofs_zipped_pagevec *vec)
{
	const unsigned ofs = offset_in_page(vec->pages);

	if (ofs < sizeof(struct page *)) {
		struct page *const prev = *vec->pages;

		DBG_BUGON(ofs);
		DBG_BUGON(vec->page == NULL);
		kunmap(vec->page);
		if ((vec->page = prev) == NULL)
			return NULL;
		vec->pages = (struct page **)(kmap(prev) + PAGE_SIZE) - 1;
	}
	return *vec->pages--;
}

static inline void z_erofs_zipped_pagevec_end(
	struct z_erofs_zipped_pagevec *vec)
{
	if (vec->page != NULL)
		kunmap(vec->page);
}

#define DE_PAGEVEC_INLINE_ENTRYS      4
/*
 * page_start
 * |  /
 * | /
 * |/
 * +------+------+--------------------------------+
 * | .... | next |              ...               |
 * +------+------+--------------------------------+
 */
struct z_erofs_de_pagevec {
	unsigned short nr, index;
	struct page *page;
	struct page **pages;
};

static inline void z_erofs_de_pagevec_end(
	struct z_erofs_de_pagevec *vec, bool atomic)
{
	if (vec->page != NULL) {
		if (!atomic) {
			kunmap(vec->page);
			return;
		}
		kunmap_atomic(vec->pages);
	}
}

/* the mid page is safe and avoid partial pages */
static inline struct page *__find_midpage(struct page **pages)
{
	struct page *const a = pages[0],
		*const b = pages[1], *const c = pages[2];

	DBG_BUGON(a == NULL);
	DBG_BUGON(b == NULL);
	DBG_BUGON(c == NULL);
	return a->index > b->index ?
		(c->index > a->index ? a : (b->index > c->index ? b : c)) :
		(c->index > b->index ? b : (a->index > c->index ? a : c));
}

static inline void __z_erofs_de_pagevec_pagedown(
	struct z_erofs_de_pagevec *vec)
{
	struct page *selected = __find_midpage(vec->pages);
	z_erofs_de_pagevec_end(vec, false);
	vec->page = selected;
	vec->pages = kmap(vec->page);
	vec->nr = PAGE_SIZE / sizeof(struct page *);
	vec->index = 0;
}

static inline void z_erofs_de_pagevec_init(
	struct z_erofs_de_pagevec *vec,
	struct page **inlinepages, unsigned i)
{
	vec->nr = DE_PAGEVEC_INLINE_ENTRYS;
	vec->page = NULL;
	vec->pages = inlinepages;

	if (i >= DE_PAGEVEC_INLINE_ENTRYS) {
		i -= DE_PAGEVEC_INLINE_ENTRYS;
		__z_erofs_de_pagevec_pagedown(vec);
		while(i > vec->nr) {
			i -= vec->nr;
			__z_erofs_de_pagevec_pagedown(vec);
		}
	}
	vec->index = i;
}

static inline void z_erofs_de_pagevec_enqueue(
	struct z_erofs_de_pagevec *vec,
	struct page *page)
{
	if (vec->index >= vec->nr)
		__z_erofs_de_pagevec_pagedown(vec);

	vec->pages[vec->index++] = page;
	debugln("%s, vec->pages[%d] = %p", __func__,
		vec->index - 1, vec->pages[vec->index - 1]);
}

static inline struct page *z_erofs_de_pagevec_dequeue(
	struct z_erofs_de_pagevec *vec)
{
	if (vec->index >= vec->nr) {
		struct page *par = vec->page;

		vec->page = __find_midpage(vec->pages);

		if (par != NULL)
			kunmap_atomic(vec->pages);
		vec->pages = kmap_atomic(vec->page);
		vec->nr = PAGE_SIZE / sizeof(struct page *);
		vec->index = 0;
	}

	return vec->pages[vec->index++];
}

#endif

