/* SPDX-License-Identifier: GPL-2.0+
 *
 * linux/fs/erofs/unzip.h
 *
 * Copyright (c) 2018 HUAWEI, Inc.
 *             http://www.huawei.com/
 * Created by Gao Xiang <gaoxiang25@huawei.com>
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file COPYING in the main directory of the Linux
 * distribution for more details.
 */
#ifndef __EROFS_UNZIP_H
#define __EROFS_UNZIP_H

#include "pagevec.h"
#include <linux/workqueue.h>

#define Z_EROFS_PACK_SIZE       (sizeof(struct z_erofs_vle_zipped_pack))

#define Z_EROFS_PACK_INITIALIZED        0x0001
#define Z_EROFS_PACK_ZIPPED             0x0002

struct z_erofs_vle_zipped_pack {
	struct erofs_page_bundle bundle;
	struct mutex lock;

	struct work_struct work;
	/*
	 * a small number of inline pages (should be > 3, currently == 4)
	 * the rest pages are dynamically managed in de_pagevec (see pagevec.h)
	 */
	struct page *pages[DE_PAGEVEC_INLINE_ENTRYS];

	unsigned flags;
	unsigned short queued_pages;
	unsigned short nr_pages;

	/* logical extent information */
	unsigned llen;
	unsigned long long la;
};

struct z_erofs_vle_zipped_iter {
	struct z_erofs_vle_zipped_pack *zip;
	struct z_erofs_de_pagevec d_pvec;

	bool already;
};

struct z_erofs_pack_info {
	struct page **pages;
	unsigned short queued_pages;
	unsigned short nr_pages;
};

#define z_erofs_vle_zipped_list_entry(zip) (&(zip)->work.entry)
#define z_erofs_vle_zipped_protect_list_entry(zip) \
	(!test_and_set_bit(WORK_STRUCT_PENDING_BIT, \
work_data_bits(&(zip)->work)))
#define z_erofs_vle_zipped_unprotect_list_entry(zip) \
	clear_bit(WORK_STRUCT_PENDING_BIT, \
work_data_bits(&(zip)->work))

struct z_erofs_onlinepage_info {
	atomic_t pending_packs;
};

static inline void z_erofs_onlinepage_init(struct page *page)
{
	struct z_erofs_onlinepage_info opi = {
		.pending_packs = ATOMIC_INIT(0)
	};

	set_page_private(page, *(unsigned long *)&opi);
	smp_wmb();
	SetPagePrivate(page);
}

static inline void z_erofs_onlinepage_setup(struct page *page, int packs)
{
	struct z_erofs_onlinepage_info *opi;
	int ret;

	DBG_BUGON(!PagePrivate(page));
	opi = (void *)&page_private(page);

	ret = atomic_add_return(packs, &opi->pending_packs);
	if (!ret) {
		ClearPagePrivate(page);
		if (!PageError(page))
			SetPageUptodate(page);
		unlock_page(page);
	}
	debugln("%s, page %p pending_packs %d count: %u",
		__func__, page, atomic_read(&opi->pending_packs),
#if (LINUX_VERSION_CODE < KERNEL_VERSION(4, 7, 0))
		atomic_read(&page->_count));
#else
		atomic_read(&page->_refcount));
#endif
}

#define z_erofs_onlinepage_endio(page)	\
	z_erofs_onlinepage_setup(page, -1)

/* unzip_generic.c */
int erofs_unzip_lz4(void *in, void *out, size_t inlen, size_t outlen);

extern void z_erofs_plain_copy(struct z_erofs_pack_info *,
	struct page **, unsigned, unsigned long long);
extern int z_erofs_unzip_generic(struct z_erofs_pack_info *,
	void *, size_t, unsigned long long, unsigned);
extern int erofs_try_to_free_vle_zipped_page(struct page *);

#include "unzip_vle.h"

#endif

